@@ -43,6 +43,8 @@ using namespace type_object_abbrev;
 
 
 const int FRAME_READY = SDL_USEREVENT;
+const int RENDER_CHUNK_SIZE = 32;
+const int DEFAULT_SPECULAR_EXP = 8;
 
 
 class already_running_error : public std::exception {
@@ -118,7 +120,7 @@ struct renderer {
     std::vector<SDL_Thread*> workers;
     scene *sc;
     SDL_Surface *destination;
-    std::atomic<unsigned int> line;
+    std::atomic<unsigned int> chunk;
     PyObject *self;
     volatile enum {NORMAL,CANCEL,QUIT} state;
 
@@ -277,11 +279,6 @@ void draw_pixel(const scene *sc,byte *&dest,const SDL_Surface *surface,int x,int
 int worker(obj_Renderer *self) {
     renderer &r = self->base;
     
-    // to minimize locking, each line is buffered
-    std::unique_ptr<byte[]> buffer;
-    int buffer_size = -1;
-
-    
     if(!r.busy_threads) {
         scoped_lock lock(r.mut);
         
@@ -294,28 +291,26 @@ int worker(obj_Renderer *self) {
     }
 
     for(;;) {
-        int byte_width = r.destination->w * r.destination->format->BytesPerPixel;
-        if(buffer_size < byte_width) {
-            buffer.reset(new byte[byte_width]);
-            buffer_size = byte_width;
-        }
+        int chunks_x = (r.destination->w + RENDER_CHUNK_SIZE - 1) / RENDER_CHUNK_SIZE;
+        int chunks_y = (r.destination->h + RENDER_CHUNK_SIZE - 1) / RENDER_CHUNK_SIZE;
 
-        int y;
-        while((y = r.line.fetch_add(1)) < r.destination->h) {
-            byte *pixels = buffer.get();
-
-            for(int x = 0; x < r.destination->w; ++x) {
-                if(UNLIKELY(r.state != renderer::NORMAL)) goto finish;
-
-                draw_pixel(r.sc,pixels,r.destination,x,y);
+        for(;;) {
+            int chunk = r.chunk.fetch_add(1);
+            int start_y = chunk/chunks_x;
+            int start_x = chunk%chunks_x;
+            if(start_y > chunks_y) break;
+            start_x *= RENDER_CHUNK_SIZE;
+            start_y *= RENDER_CHUNK_SIZE;
+            
+            for(int y = start_y; y < std::min(start_y+RENDER_CHUNK_SIZE,r.destination->h); ++y) {
+                byte *pixels = reinterpret_cast<byte*>(r.destination->pixels) + y * r.destination->pitch + start_x * r.destination->format->BytesPerPixel;
+                
+                for(int x = start_x; x < std::min(start_x+RENDER_CHUNK_SIZE,r.destination->w); ++x) {
+                    if(UNLIKELY(r.state != renderer::NORMAL)) goto finish;
+                    
+                    draw_pixel(r.sc,pixels,r.destination,x,y);
+                }
             }
-
-            if(SDL_MUSTLOCK(r.destination)) SDL_LockSurface(r.destination);
-            memcpy(
-                reinterpret_cast<byte*>(r.destination->pixels) + y * r.destination->pitch,
-                buffer.get(),
-                byte_width);
-            if(SDL_MUSTLOCK(r.destination)) SDL_UnlockSurface(r.destination);
         }
 
     finish:
@@ -347,6 +342,7 @@ int worker(obj_Renderer *self) {
                        workers is harmless. */
                     r.barrier.signal_all();
                 }
+                if(SDL_MUSTLOCK(r.destination)) SDL_UnlockSurface(r.destination);
                 r.sc->unlock();
                 SDL_FreeSurface(r.destination);
                 Py_DECREF(self);
@@ -456,20 +452,27 @@ PyObject *obj_Renderer_begin_render(obj_Renderer *self,PyObject *args,PyObject *
             if(r.busy_threads) throw already_running_error();
             
             r.busy_threads = r.threads;
-            r.line.store(0,std::memory_order_relaxed);
+            r.chunk.store(0,std::memory_order_relaxed);
             ++r.job;
             r.sc = sc;
             sc->lock();
             try {
                 r.destination = PySurface_AsSurface(dest);
                 ++r.destination->refcount;
-                r.barrier.signal_all();
+                if(SDL_MUSTLOCK(r.destination)) SDL_LockSurface(r.destination);
+                try {
+                    r.barrier.signal_all();
+                } catch(...) {
+                    if(SDL_MUSTLOCK(r.destination)) SDL_UnlockSurface(r.destination);
+                    throw;
+                }
             } catch(...) {
                 r.sc->unlock();
                 throw;
             }
         } catch(...) {
             Py_DECREF(self);
+            throw;
         }
 
         Py_RETURN_NONE;
@@ -811,90 +814,94 @@ PyTypeObject color_obj_base::_pytype = make_type_object(
     tp_new = &obj_Color_new);
 
 
+struct py_mem_deleter {
+    void operator()(char *x) {
+        PyMem_Free(x);
+    }
+};
+std::unique_ptr<char,py_mem_deleter> f_to_s(float x) {
+    char *r = PyOS_double_to_string(x,'r',0,0,nullptr);
+    if(!r) throw py_error_set();
+    return std::unique_ptr<char,py_mem_deleter>(r);
+}
+
 PyObject *obj_Material_repr(material *self) {
-    PyObject *ret = nullptr;
-    char *cr;
-    char *cg = nullptr;
-    char *cb = nullptr;
-    char *o = nullptr;
-    char *r = nullptr;
-    
-    if((cr = PyOS_double_to_string(self->c.r,'r',0,0,nullptr))) {
-        if((cg = PyOS_double_to_string(self->c.g,'r',0,0,nullptr))) {
-            if((cb = PyOS_double_to_string(self->c.b,'r',0,0,nullptr))) {
-                if(self->reflectivity == 0 && self->opacity == 1) {
-                    ret = PYSTR(FromFormat)("Material((%s,%s,%s))",cr,cg,cb);
-                } else if((o = PyOS_double_to_string(self->opacity,'r',0,0,nullptr))) {
-                    if(self->reflectivity == 0) {
-                        ret = PYSTR(FromFormat)("Material((%s,%s,%s),%s)",cr,cg,cb,o);
-                    } else if((r = PyOS_double_to_string(self->reflectivity,'r',0,0,nullptr))) {
-                        ret = PYSTR(FromFormat)("Material((%s,%s,%s),%s,%s)",cr,cg,cb,o,r);
-                    }
-                }
-            }
+    try {
+        return PYSTR(FromFormat)("Material((%s,%s,%s),%s,%s,%s,%s,(%s,%s,%s))",
+            f_to_s(self->c.r).get(),
+            f_to_s(self->c.g).get(),
+            f_to_s(self->c.b).get(),
+            f_to_s(self->opacity).get(),
+            f_to_s(self->reflectivity).get(),
+            f_to_s(self->specular_intensity).get(),
+            f_to_s(self->specular_exp).get(),
+            f_to_s(self->specular.r).get(),
+            f_to_s(self->specular.g).get(),
+            f_to_s(self->specular.b).get());
+    } PY_EXCEPT_HANDLERS(nullptr)
+}
+
+void read_color(color &to,PyObject *from,const char *field) {
+    if(PyTuple_Check(from)) {
+        if(PyTuple_GET_SIZE(from) != 3) {
+            PyErr_Format(PyExc_ValueError,"\"%s\" must have exactly 3 values",field);
+            throw py_error_set();
         }
+        to.r = from_pyobject<float>(PyTuple_GET_ITEM(from,0));
+        to.g = from_pyobject<float>(PyTuple_GET_ITEM(from,1));
+        to.b = from_pyobject<float>(PyTuple_GET_ITEM(from,2));
+    } else {
+        to = get_base<color>(from);
     }
-    
-    PyMem_Free(cr);
-    PyMem_Free(cg);
-    PyMem_Free(cb);
-    PyMem_Free(o);
-    PyMem_Free(r);
-    
-    return ret;
 }
 
 PyObject *obj_Material_new(PyTypeObject *type,PyObject *args,PyObject *kwds) {
     try {
-        const char *names[] = {"color","opacity","reflectivity"};
+        const char *names[] = {"color","opacity","reflectivity","specular_intensity","specular_exp","specular_color"};
         get_arg ga(args,kwds,names,"Material.__new__");
         auto obj_c = ga(true);
         float o = 1;
         float r = 0;
+        float si = 1;
+        float se = DEFAULT_SPECULAR_EXP;
         auto temp = ga(false);
         if(temp) o = from_pyobject<float>(temp);
         temp = ga(false);
         if(temp) r = from_pyobject<float>(temp);
+        temp = ga(false);
+        if(temp) si = from_pyobject<float>(temp);
+        temp = ga(false);
+        if(temp) se = from_pyobject<float>(temp);
+        auto obj_s = ga(false);
         ga.finished();
         
         auto ptr = py::check_obj(type->tp_alloc(type,0));
         auto base = reinterpret_cast<material*>(ptr);
         
-        if(PyTuple_Check(obj_c)) {
-            if(PyTuple_GET_SIZE(obj_c) != 3) {
-                PyErr_SetString(PyExc_ValueError,"\"color\" must have exactly 3 values");
-                return nullptr;
-            }
-            base->c.r = from_pyobject<float>(PyTuple_GET_ITEM(obj_c,0));
-            base->c.g = from_pyobject<float>(PyTuple_GET_ITEM(obj_c,1));
-            base->c.b = from_pyobject<float>(PyTuple_GET_ITEM(obj_c,2));
-        } else {
-            base->c = get_base<color>(obj_c);
-        }
+        read_color(base->c,obj_c,"color");
+        if(obj_s) read_color(base->specular,obj_s,"specular_color");
+        else base->specular = color(1,1,1);
         
         base->opacity = o;
         base->reflectivity = r;
+        base->specular_intensity = si;
+        base->specular_exp = se;
         
         return ptr;
     } PY_EXCEPT_HANDLERS(nullptr)
 }
 
-int obj_Material_set_color(material *self,PyObject *arg,void*) {
-    try {
-        setter_no_delete(arg);
-        self->c = get_base<color>(arg);
-        return 0;
-    } PY_EXCEPT_HANDLERS(-1)
-}
-
 PyGetSetDef obj_Material_getset[] = {
-    {const_cast<char*>("color"),OBJ_GETTER(material,self->c),reinterpret_cast<setter>(&obj_Material_set_color),NULL,NULL},
+    {const_cast<char*>("color"),OBJ_GETTER(material,self->c),OBJ_BASE_SETTER(material,self->c),NULL,NULL},
+    {const_cast<char*>("specular"),OBJ_GETTER(material,self->specular),OBJ_BASE_SETTER(material,self->specular),NULL,NULL},
     {NULL}
 };
 
 PyMemberDef obj_Material_members[] = {
     {const_cast<char*>("opacity"),T_FLOAT,offsetof(material,opacity),0,NULL},
     {const_cast<char*>("reflectivity"),T_FLOAT,offsetof(material,reflectivity),0,NULL},
+    {const_cast<char*>("specular_intensity"),T_FLOAT,offsetof(material,reflectivity),0,NULL},
+    {const_cast<char*>("specular_exp"),T_FLOAT,offsetof(material,reflectivity),0,NULL},
     {NULL}
 };
 
